# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Verification of the plan."""

import json
from pkg_resources import resource_filename as resource
from jsonschema import validate
from jsonschema.exceptions import ValidationError

SCHEMA_FILE_NAME = resource('dsplab', 'data/plan-schema.json')


class VerifyError(Exception):
    """Verification error."""


def check_plan(plan_dict):
    """Check plan's dictionary."""

    schema = _load_schema(SCHEMA_FILE_NAME)
    _check_plan_schema(plan_dict, schema)

    ids = _get_ids(plan_dict)

    for node in plan_dict["nodes"]:
        _check_node(node, ids)

    _check_plan_inputs(plan_dict, ids)
    _check_plan_outputs(plan_dict, ids)


#
# plan

def _check_plan_schema(plan_dict, schema):
    try:
        validate(instance=plan_dict, schema=schema)

    except ValidationError as ex:
        raise VerifyError from ex


def _check_plan_inputs(plan_dict, ids):
    for inp in _get_plan_inputs(plan_dict):
        if inp not in ids:
            raise VerifyError("Unknown plan input: {}".format(inp))


def _check_plan_outputs(plan_dict, ids):
    for out in _get_plan_outputs(plan_dict):
        if out not in ids:
            raise VerifyError("Unknown plan output: {}".format(out))


def _get_ids(plan_dict):
    ids = {}
    for node in plan_dict["nodes"]:
        node_id = node["id"]
        if node_id in ids:
            raise VerifyError("Duplicated ID: {}".format(node_id))
        ids[node_id] = node
    return ids


def _get_plan_inputs(plan_dict):
    return _get_value_or_list(plan_dict, "inputs")


def _get_plan_outputs(plan_dict):
    return _get_value_or_list(plan_dict, "outputs")


#
# node

def _check_node(node_dict, ids):
    _check_node_inputs(node_dict, ids)

    if _get_node_class(node_dict) in ["WorkNode", "MapNode"]:
        if "work" not in node_dict:
            raise VerifyError("No work in node {}".format(node_dict["id"]))


def _check_node_inputs(node_dict, ids):
    node_id = node_dict["id"]
    for inp in _get_node_inputs(node_dict):
        if inp not in ids:
            raise VerifyError("Unknown input {} in node {}".
                              format(inp, node_id))

        if inp == node_id:
            raise VerifyError("Node {} uses itself as input".format(inp))


def _get_node_inputs(node_dict):
    return _get_value_or_list(node_dict, "inputs")


def _get_node_class(node_dict):
    if "class" in node_dict:
        return node_dict["class"]
    return "WorkNode"


#
# common

def _load_schema(file_name):
    with open(file_name) as buf:
        data = json.load(buf)

    return data


def _get_value_or_list(dct, key):
    if key not in dct:
        return []
    return dct[key]



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
from dsplab import modulation as mod


class Test_harm(unittest.TestCase):
    def test_amplitude_and_rate(self):
        rate = 50
        length = 120
        amp = 2.5

        xs, ts = mod.harm(length=length, sample_rate=rate, amp=amp, freq=1)
        self.assertEqual(len(xs), len(ts))
        self.assertEqual(len(xs), length*rate)
        self.assertAlmostEqual(max(xs), amp)
        self.assertAlmostEqual(min(xs), -amp)
        self.assertAlmostEqual(sum(xs), 0)
        self.assertAlmostEqual(ts[1]-ts[0], 1/rate)

    def test_frequency(self):
        xs = mod.harm(length=60, sample_rate=1, amp=1, freq=1)[0]
        self.assertAlmostEqual(sum(xs), 60)

    def test_noised_amplitude(self):
        xs = mod.harm(length=60, sample_rate=50, amp=1, freq=1,
                      noise_amp=lambda: 1)[0]
        self.assertAlmostEqual(max(xs), 2)
        self.assertAlmostEqual(min(xs), -2)

    def test_noised_phase(self):
        xs = mod.harm(length=60, sample_rate=1, amp=1, freq=1,
                      noise_ph=lambda: np.pi/2)[0]
        self.assertAlmostEqual(sum(xs), 0)
        self.assertAlmostEqual(max(xs), 0)
        self.assertAlmostEqual(min(xs), 0)


class Test_amp_mod(unittest.TestCase):
    def test_empty(self):
        xs, ts = mod.amp_mod(0, 50, lambda t: 1, 10)
        self.assertEqual(len(xs), 0)
        self.assertEqual(len(ts), 0)

    def test_result_len(self):
        xs, ts = mod.amp_mod(1, 50, lambda t: 1, 10)
        self.assertEqual(len(xs), 50)
        self.assertEqual(len(ts), 50)

    def test_noise_amp(self):
        xs = mod.amp_mod(1, 50, lambda t: 1, 1, noise_amp=lambda: 1)[0]
        self.assertEqual(max(xs), 2)
        self.assertEqual(min(xs), 0)

    def test_noise_ph_touch(self):
        xs = mod.amp_mod(1, 50, lambda t: 1, 1, noise_ph=lambda: 0)[0]
        self.assertEqual(len(xs), 50)


class Test_freq_mod(unittest.TestCase):
    def test_touch(self):
        xs, phs, ts = mod.freq_mod(1, 50, 1, lambda t: 10,
                                   noise_amp=lambda: 1, noise_ph=lambda: 0)
        self.assertEqual(len(xs), 50)
        self.assertEqual(len(phs), 50)
        self.assertEqual(len(ts), 50)


class Test_phase_mod(unittest.TestCase):
    def test_touch(self):
        xs, ts = mod.phase_mod(1, 50, 1, 10, lambda t: 10,
                               noise_amp=lambda: 1, noise_ph=lambda: 0)
        self.assertEqual(len(xs), 50)
        self.assertEqual(len(ts), 50)


class Test_freq_amp_mod(unittest.TestCase):
    def test_touch(self):
        xs, phs, ts = mod.freq_amp_mod(1, 50, lambda t: 1, lambda t: 10)
        self.assertEqual(len(xs), 50)
        self.assertEqual(len(phs), 50)
        self.assertEqual(len(ts), 50)


class Test_envelope_by_extremums(unittest.TestCase):
    def test_result_len(self):
        x = np.array([0, 1, 0, -1, 0, 1, 0, -1, 0, 1, 0])
        e, t = mod.envelope_by_extremums(x)
        self.assertEqual(len(e) + len(t), 5 + 5)


class Test_digital_hilbert_filter(unittest.TestCase):
    def test_filter_len(self):
        self.assertEqual(len(mod.digital_hilbert_filter(3)), 3)


class Test_linint(unittest.TestCase):
    def test_empty(self):
        t = np.array([])
        x = np.array([])
        t_new = np.array([])
        x_new = mod.linint(x, t, t_new)
        self.assertEqual(len(x_new), 0)

    def test_not_empty(self):
        t = np.array([0, 1, 3, 4, 6, 7, 9])
        x = np.array([0, 1, 3, 4, 6, 7, 9])
        t_new = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        x_new = mod.linint(x, t, t_new)
        self.assertEqual(sum(x_new), 45)

    def test_nans(self):
        t = np.array([0, 1, 3, 4, 6, 7])
        x = np.array([0, 1, 3, 4, 6, 7])
        t_new = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        x_new = mod.linint(x, t, t_new)
        self.assertEqual(np.nansum(x_new), 28)
        self.assertEqual(len(x_new), 10)


class Test_wave_lens(unittest.TestCase):
    def test_touch(self):
        t = np.array([0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5])
        x = np.array([1, -1,  1, -1,  1, -1,  1, -1,  1, -1,  1])
        wl, t_wl = mod.wave_lens(x, t)
        self.assertEqual(len(wl), 9)
        self.assertEqual(len(t_wl), 9)
        self.assertEqual(sum(wl), 9)
        self.assertEqual(sum(t_wl), 27)


class Test_freqs_by_wave_len(unittest.TestCase):
    def test_touch(self):
        t = np.array([0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5])
        x = np.array([1, -1,  1, -1,  1, -1,  1, -1,  1, -1,  1])
        f, t_f = mod.freqs_by_wave_len(x, t)
        self.assertEqual(len(f), 9)
        self.assertEqual(len(t_f), 9)
        self.assertEqual(np.sum(f), 9)


class Test_freq_by_zeros(unittest.TestCase):
    def test_touch(self):
        x = np.array([1, 0, -1,  0])
        self.assertAlmostEqual(mod.freq_by_zeros(x, 4), 1)


class TestFreqByExtremums(unittest.TestCase):
    def test_short_signal(self):
        with self.assertRaises(ValueError):
            mod.freq_by_extremums([1, 0], 4)

    def test_single_wave(self):
        x = np.array([1, 0, -1,  0])
        self.assertAlmostEqual(mod.freq_by_extremums(x, 4), 1)



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from dsplab.prony import prony_decomp


class TestProny(unittest.TestCase):
    def test_prony_sum_length_of_result(self):
        x = [255, -255, 128, -128, 64, -64, 32, -32, 16, -16, 8, -8,
             4, -4, 2, -2, 1, -1, 0, 0]
        L = 4
        ms, cs, es = prony_decomp(x, L)
        self.assertEqual(len(ms) + len(cs) + len(es), 4 + 4 + 4)

    def test_prony_length_of_components(self):
        x = [255, -255, 128, -128, 64, -64, 32, -32, 16, -16, 8, -8,
             4, -4, 2, -2, 1, -1, 0, 0]
        L = 4
        ms, cs, es = prony_decomp(x, L)
        self.assertEqual(len(es[0]), len(x))

    def test_prony_wrong_number_of_components(self):
        x = [255, -255, 128, -128, 64, -64, 32, -32, 16, -16, 8, -8,
             4, -4, 2, -2, 1, -1, 0, 0]
        L = len(x)//2 + 1
        res = prony_decomp(x, L)
        self.assertEqual(res, None)

    def test_prony_double_number_of_components_equal_to_n(self):
        x = [255, -255, 128, -128, 64, -64, 32, -32, 16, -16, 8, -8,
             4, -4, 2, -2, 1, -1, 0, 0]
        L = 10
        ms, cs, es = prony_decomp(x, L)
        self.assertEqual(len(ms) + len(cs) + len(es), 10 + 10 + 10)



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
from dsplab.spectran import spectrum, stft, calc_specgram


class TestSpectrum(unittest.TestCase):
    def test_spectrum_len(self):
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        X, f_X = spectrum(x, sample_rate=1)
        self.assertEqual(len(x), len(X))

    def test_spectrum_one_side_len(self):
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        X, f_X = spectrum(x, sample_rate=1, one_side=True)
        self.assertEqual(len(x)//2, len(X))

    def test_spectrum_need_expand_len(self):
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        X, f_X = spectrum(x, sample_rate=1, extra_len=16)
        self.assertEqual(16, len(X))


class TestSTFT(unittest.TestCase):
    def test_stft_number_of_spectrums_no_overlap(self):
        fs = 1
        x = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        Xs = stft(x, fs, 2, 2, window='hanning', nfft=None, padded=False)
        self.assertEqual(len(Xs), 5)

    def test_stft_number_of_spectrums_overlap(self):
        fs = 1
        x = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        Xs = stft(x, fs, 2, 1, window='hanning', nfft=None, padded=False)
        self.assertEqual(len(Xs), 9)

    def test_stft_len_of_spectrum_dont_add_zeros_to_segments(self):
        fs = 1
        x = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        Xs = stft(x, fs, 2, 2, window='hanning', nfft=None, padded=False)
        self.assertEqual(len(Xs[0]), 2)

    def test_stft_len_of_spectrum_add_zeros_to_segments(self):
        fs = 1
        x = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        Xs = stft(x, fs, 2, 2, window='hanning', nfft=4, padded=False)
        self.assertEqual(len(Xs[0]), 4)

    def test_stft_padding_true(self):
        fs = 1
        x = np.array([0, 1, 2, 3, 4, 5, 6])
        Xs = stft(x, fs, 3, 3, window='hanning', nfft=None, padded=True)
        self.assertEqual(len(Xs), 3)

    def test_stft_short_signal(self):
        x = np.linspace(0, 1, 100)
        Xs = stft(x, sample_rate=1, nseg=256, nstep=128)
        self.assertEqual(len(Xs), 0)


class TestSpecgram(unittest.TestCase):
    def test_calc_specgram_just_run(self):
        x = np.linspace(0, 100, 1024)
        calc_specgram(x)
        self.assertTrue(True)

    def test_calc_specgram_short_signal(self):
        x = np.linspace(0, 1, 100)
        Xs, t = calc_specgram(x, nseg=256)
        self.assertEqual(len(Xs), 0)
        self.assertEqual(len(t), 0)

    def test_calc_specgram_len_signal_and_nseg_are_equal(self):
        x = np.linspace(0, 1, 256)
        Xs, t = calc_specgram(x, nseg=256)
        self.assertEqual(len(Xs[0]), 1)
        self.assertEqual(len(t), 1)

    def test_calc_specgram_time_used_just_run(self):
        t = np.linspace(0, 100, 1024)
        x = np.cos(t)
        calc_specgram(x, tdata=t)
        self.assertTrue(True)

    def test_calc_specgram_bound_used_just_run(self):
        fs = 50
        t = np.arange(0, 20, 1/fs)
        x = np.cos(2*np.pi*1*t)
        calc_specgram(x, tdata=t, freq_bounds=(0.5, 1.5))
        self.assertTrue(True)

    def test_calc_specgram_fs_used_just_run(self):
        fs = 50
        t = np.arange(0, 20, 1/fs)
        x = np.cos(2*np.pi*1*t)
        calc_specgram(x, sample_rate=fs)
        self.assertTrue(True)



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from dsplab import filtration as flt


class TestHaarOneStep(unittest.TestCase):
    def test_haar_one_step_even_len(self):
        t = [0, 1, 2, 3, 4, 5]
        x = [1, 2, 3, 4, 5, 6]
        x_s, x_d, t_new = flt.haar_one_step(x, t)
        self.assertEqual(len(x_s), 3)
        self.assertEqual(len(x_d), 3)
        self.assertEqual(len(t_new), 3)

    def test_haar_one_step_sum_of_values(self):
        t = [0, 1, 2, 3, 4, 5]
        x = [1, 2, 3, 4, 5, 6]
        x_s, x_d, t_new = flt.haar_one_step(x, t, denominator=2)
        self.assertEqual(sum(x_s), 10.5)
        self.assertEqual(sum(x_d), -1.5)

    def test_haar_one_step_odd_len(self):
        t = [0, 1, 2, 3, 4, 5, 6]
        x = [1, 2, 3, 4, 5, 6, 7]
        x_s, x_d, t_new = flt.haar_one_step(x, t)
        self.assertEqual(len(x_s), 3)
        self.assertEqual(len(x_d), 3)
        self.assertEqual(len(t_new), 3)


class TestHaarScaling(unittest.TestCase):
    def test_haar_scaling_len_sum(self):
        t = [0, 1, 2, 3, 4, 5, 6, 7]
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        x_resampled, t_new = flt.haar_scaling(x, t, 3)
        self.assertEqual(len(x_resampled), 1)
        self.assertEqual(sum(x_resampled), 4.5)

    def test_haar_scaling_time(self):
        t = [0, 1, 2, 3, 4, 5, 6, 7]
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        x_resampled, t_new = flt.haar_scaling(x, t, 3)
        self.assertEqual(len(t_new), 1)
        self.assertEqual(t_new[0], 7)


class TestOrder(unittest.TestCase):
    def test_find_butt_bandpass_order_1(self):
        band = (1.93, 2.14)
        sampling_steps = 2
        fs = 50/2**sampling_steps
        order = flt.find_butt_bandpass_order(band, fs)
        self.assertTrue(order > 5)

    def test_find_butt_bandpass_order_2(self):
        band = (3.84, 4)
        sampling_steps = 2
        fs = 50/2**sampling_steps
        order = flt.find_butt_bandpass_order(band, fs)
        self.assertTrue(order > 5)


class TestStupidFilters(unittest.TestCase):
    def test_stupid_lowpass_filter_just_run(self):
        x = [1, 2, 3, 4, 5, 6]
        flt.stupid_lowpass_filter(x, sample_rate=1, cutoff=0.25)
        self.assertTrue(True)

    def test_stupid_bandpass_filter_just_run(self):
        x = [1, 2, 3, 4, 5, 6]
        flt.stupid_bandpass_filter(x, sample_rate=1, bandpass=(0.1, 0.2))
        self.assertTrue(True)


class TestTrend(unittest.TestCase):
    def test_trend_smooth_just_run(self):
        x = [1, 2, 3, 4, 5, 6, 7, 8]
        flt.trend_smooth(x)
        self.assertTrue(True)



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Dsplab is the package for building generalized signal processing schemes.

In a different subject areas, data analysis tasks exist in which a
certain general structure is distinguished, while the number of
possible methods for implementing subtasks increases. The question
arises: which combination of methods is better? Dsplab allows you to
describe the overall scheme of the task (plan) and assign the
so-called "workers" to perform the subtasks. So the main field of
application of dsplab is the investigation of variety of methods
solving the same signal processing task.

"""

from setuptools import setup
from dsplab import __version__


DOCLINES = __doc__.split("\n")


INSTALL_REQUIRES = [
    'numpy>=1.14',
    'scipy>=0.19',
    'jsonschema>=3.2',
]


PACKAGES = [
    'dsplab',
    'dsplab.flow',
]


PACKAGE_DATA = {
    'dsplab': [
        'data/*',
    ],
}


setup(
    name='dsplab',
    version=__version__,
    description=DOCLINES[0],
    author="Aleksandr Popov, Kirill Butin",
    author_email="aleneus@gmail.com",
    license="LGPLv3",
    keywords="digital signal processing",
    url="https://github.com/aleneus/dsplab",
    long_description="\n".join(DOCLINES[2:]),
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
)



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Functions for modulation and demodulation."""

from warnings import warn
from math import pi, cos, isnan
import numpy as np
from numpy import unwrap, angle, diff
import scipy.signal as sig


def harm(length, sample_rate, amp, freq, phi=0,
         noise_ph=None, noise_amp=None,
         noise_a=None, noise_f=None):
    """Generate harmonic signal.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    freq: Object
        Frequency of signal (Hz).
    phi: float
        Initial phase (radians).
    noise_ph: callable
        Returns noise value added to full phase.
    noise_amp: callable
        Returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """
    _noise_amp = _arg_new_or_depr(noise_a, noise_amp, "noise_a", "noise_amp")
    _noise_ph = _arg_new_or_depr(noise_f, noise_ph, "noise_f", "noise_ph")

    ts = np.arange(0, length, 1 / sample_rate)

    xs = []
    for t in ts:
        x = _ns(amp, _noise_amp) * cos(_ns(2*pi*freq*t + phi, _noise_ph))
        xs.append(x)

    return np.array(xs), ts


def amp_mod(length, sample_rate, func, freq, phi=0,
            noise_ph=None, noise_amp=None,
            noise_f=None, noise_a=None):
    """Amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    freq: float
        Frequency of signal (Hz).
    phi: float
        Initial phase (radians).
    func: Object
        Function that returns amplitude value depending on time.
    noise_ph: callable
        Returns noise value added to full phase.
    noise_amp: callable
        Returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """
    _noise_amp = _arg_new_or_depr(noise_a, noise_amp, "noise_a", "noise_amp")
    _noise_ph = _arg_new_or_depr(noise_f, noise_ph, "noise_f", "noise_ph")

    ts = np.arange(0, length, 1/sample_rate)

    full_phase = phi
    delta_ph = 2*pi*freq / sample_rate
    xs = []
    for t in ts:
        x = _ns(func(t) * cos(_ns(full_phase, _noise_ph)), _noise_amp)
        xs.append(x)
        full_phase += delta_ph

    return np.array(xs), ts


def freq_mod(length, sample_rate, amp, func, phi=0,
             noise_ph=None, noise_amp=None,
             noise_f=None, noise_a=None):
    """Amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    phi: float
        Initial phase (radians).
    func: Object
        Function that returns frequency values (in Hz) depending on
        time.
    noise_ph: callable
        Returns noise value added to full phase.
    noise_amp: callable
        Returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Full phase values.
    : np.array
        Time values.
    """
    _noise_amp = _arg_new_or_depr(noise_a, noise_amp, "noise_a", "noise_amp")
    _noise_ph = _arg_new_or_depr(noise_f, noise_ph, "noise_f", "noise_ph")

    ts = np.arange(0, length, 1/sample_rate)

    full_phase = phi
    xs, phs = [], []
    for t in ts:
        x = _ns(amp * cos(_ns(full_phase, _noise_ph)), _noise_amp)
        xs.append(x)
        phs.append(full_phase)
        full_phase += 2*pi*func(t) / sample_rate

    return np.array(xs), np.array(phs), ts


def phase_mod(length, sample_rate, amp, freq, func,
              noise_ph=None, noise_amp=None,
              noise_f=None, noise_a=None):
    """Phase modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    amp: float
        Amplitude of signal.
    freq: float
        Frequency of signal (Hz).
    func: Object
        Function that returns phase values (in radians) depending on
        time.
    noise_ph: callable
        Returns noise value added to full phase.
    noise_amp: callable
        Returns noise value added to amplitude.

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Time values.
    """
    _noise_amp = _arg_new_or_depr(noise_a, noise_amp, "noise_a", "noise_amp")
    _noise_ph = _arg_new_or_depr(noise_f, noise_ph, "noise_f", "noise_ph")

    ts = np.arange(0, length, 1/sample_rate)
    xs = []
    for t in ts:
        arg = _ns(2*pi*freq*t + func(t), _noise_ph)
        x = _ns(amp * cos(arg), _noise_amp)
        xs.append(x)

    return np.array(xs), ts


def freq_amp_mod(length, sample_rate, a_func, f_func, phi=0):
    """Simultaneous frequency and amplitude modulation.

    Parameters
    ----------
    length: float
        Length pf signal (sec).
    sample_rate: float
        Sampling frequency (Hz).
    a_func: Object
        Function that returns amplitude value depending on time.
    f_func: Object
        Function that returns frequency values (in Hz) depending on
        time.
    phi: float
        Initial phase (radians).

    Returns
    -------
    : np.array
        Signal values.
    : np.array
        Full phase values.
    : np.array
        Time values.
    """
    ts = np.arange(0, length, 1/sample_rate)

    full_phase = phi
    xs = []
    phs = []
    for t in ts:
        xs.append(a_func(t) * cos(full_phase))
        phs.append(full_phase)
        full_phase += 2*pi * f_func(t) / sample_rate

    return np.array(xs), np.array(phs), ts


def iq_demod(xdata, tdata, f_central, a_coeffs, b_coeffs):
    """Return instantaneous frequency of modulated signal using IQ
    processing.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    tdata: array_like
        Time values.
    f_central: float
        Carrier frequency.
    a_coeffs: array_like
        a values of filter.
    b_coeffs: array_like
        b values of filter.

    Returns
    -------
    : np.ndarray of floats
        Instantaneous frequency values.
    : np.ndarray
        Time values.
    """
    muli = xdata * np.cos(2*pi * f_central * tdata)
    mulq = xdata * np.sin(2*pi * f_central * tdata)
    muli_low = sig.lfilter(b_coeffs, a_coeffs, muli)
    mulq_low = sig.lfilter(b_coeffs, a_coeffs, mulq)
    analytic = muli_low + 1j * mulq_low
    phase = -unwrap(angle(analytic))
    freq = diff(phase) / 2 / pi / (tdata[1] - tdata[0]) + f_central
    return freq, tdata[:-1]


def digital_hilbert_filter(ntaps=101, window='hamming'):
    """Calculate digital hilbert filter.

    Parameters
    ----------
    ntaps: integer
        Length of filter.
    window: str
        Window. Default is 'hamming'.

    Returns
    -------
    : np.array
        Filter.
    """
    if ntaps % 2 == 0:
        raise ValueError("ntaps of digital Hilbert filter must be odd.")

    coeffs = np.zeros(ntaps)
    num = ntaps // 2
    for k in range(1, num+1, 2):
        coeffs[num + k] = 2 / pi / k
        coeffs[num - k] = -2 / pi / k

    wind = sig.get_window(window, ntaps)
    coeffs *= wind
    return coeffs


def envelope_by_extremums(xdata, sample_rate=1, tdata=None):
    """Calculate envelope by local extremums of signal.

    Parameters
    ----------
    xdata: array_like
        Signal values.
    sample_rate: float
        Sampling frequency.
    tdata: array_like
        Time values. Use it for unregular discretized input signal.

    Returns
    --------
    : np.array
        Damping values.
    : np.array
        Time values.
    """
    if tdata is None:
        tdata = np.linspace(0, (len(xdata)-1)/sample_rate, len(xdata))
    t_new = []
    x_new = []
    xabs = abs(xdata)
    for x_l, x_c, x_r, t_c in zip(xabs[:-2], xabs[1:-1],
                                  xabs[2:], tdata[1:-1]):
        if (x_l < x_c) and (x_c >= x_r):
            t_new.append(t_c)
            x_new.append(x_c)
    if xabs[-1] > xabs[-2]:
        t_new.append(tdata[-1])
        x_new.append(xabs[-1])
    return np.array(x_new), np.array(t_new)


def freq_by_extremums(xdata, sample_rate):
    """Calculate frequency of oscillating signal by extremums.

    Parameters
    ----------
    xdata: array_like
        Values of input signals.
    sample_rate: float
        Sampling frequency (Hz).

    Returns
    -------
    : float
        Frequency.
    """

    if len(xdata) < 3:
        raise ValueError("Short signal")

    T = len(xdata) / sample_rate

    max_total = 0
    min_total = 0
    for prev, curr, nxt in zip(xdata[:-2], xdata[1:-1], xdata[2:]):
        if (prev < curr) and (curr >= nxt):
            max_total += 1
        if (prev > curr) and (curr <= nxt):
            min_total += 1

    if xdata[0] > xdata[1]:
        max_total += 1

    if xdata[0] < xdata[1]:
        min_total += 1

    return (max_total + min_total) / 2 / T


def freq_by_zeros(xdata, sample_rate):
    """Calculate average frequency of detrended oscillating signal by
    counting zeros."""

    T = len(xdata) / sample_rate

    zeros_total = 0
    for prev, curr in zip(xdata[:-1], xdata[1:]):
        if prev * curr < 0:
            zeros_total += 1
        elif prev != 0 and curr == 0:
            zeros_total += 1

    return zeros_total / 2 / T


def wave_lens(xdata, tdata):
    """Calculate wave lengths of signal by space between zeros.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.

    Returns
    -------
    : np.ndarray
        Wave lengths.
    : np.ndarray
        Time values.
    """
    tms = []
    for t_c, x_p, x_c in zip(tdata[1:], xdata[:-1], xdata[1:]):
        if x_p * x_c < 0:
            tms.append(t_c)
    lens = np.diff(tms) * 2
    t_lens = np.array(tms[1:])
    return lens, t_lens


def freqs_by_wave_len(xdata, tdata, cut_nans=True):
    """Calculate frequencies using lenghs of waves and linear
    interpolation.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.
    cut_nans: bool
        If True, the nan values at the ends of the of the produced
        array will removed.

    Returns
    -------
    : np.ndarray
        Freqs values.
    """
    wls, t_wl = wave_lens(xdata, tdata)
    freqs = 1 / linint(wls, t_wl, tdata)

    if cut_nans:
        freqs_cut = []
        t_cut = []
        for f, t in zip(freqs, tdata):
            if f is not None and not isnan(f):
                freqs_cut.append(f)
                t_cut.append(t)
        return np.array(freqs_cut), t_cut

    return freqs


def linint(xdata, tdata, ts_new):
    """Find values of xdata in ts_new points.

    Parameters
    ----------
    xdata: np.ndarray
        Signal values.
    tdata: np.ndarray
        Time values.
    ts_new: np.ndarray
        New time values.

    Returns
    -------
    : np.ndarray
        New signal values.
    """
    x_new = np.zeros(len(ts_new)) * np.nan
    for x_p, t_p, x_c, t_c in zip(xdata[:-1], tdata[:-1],
                                  xdata[1:], tdata[1:]):
        k = (x_c - x_p) / (t_c - t_p)
        b = x_p - k*t_p
        ind = (ts_new >= t_p) & (ts_new <= t_c)
        x_new[ind] = k * ts_new[ind] + b
    return x_new


def _ns(x, func=None):
    if func:
        return x + func()
    return x


def _arg_new_or_depr(depr, new, depr_name, new_name):
    if depr is not None:
        msg = "{} is deprecated, use {}".format(depr_name, new_name)
        warn(msg, FutureWarning)

    if new is not None:
        return new

    return depr



# Copyright (C) 2017-2021 Aleksandr Popov
# Copyright (C) 2021 Kirill Butin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
from dsplab.helpers import is_iterable


class Test_is_iterable(unittest.TestCase):
    def test_list(self):
        self.assertTrue(is_iterable([1, 2, 3]))

    def test_single_value(self):
        self.assertFalse(is_iterable(1))

    def test_none(self):
        self.assertFalse(is_iterable(None))

    def test_nparray(self):
        self.assertTrue(is_iterable(np.array([1, 2, 3])))
